@@ -26,9 +26,11 @@ import java.util.List;
 import java.util.SortedMap;
 import java.util.TreeMap;
 import java.util.concurrent.CompletableFuture;
+import java.util.concurrent.CopyOnWriteArrayList;
 import java.util.concurrent.Executors;
 import java.util.concurrent.ScheduledExecutorService;
 import java.util.concurrent.TimeUnit;
+import java.util.concurrent.atomic.AtomicInteger;
 import java.util.concurrent.atomic.AtomicLong;
 import java.util.function.BiFunction;
 import java.util.stream.Collectors;
@@ -45,6 +47,7 @@ import org.apache.pulsar.client.api.MessageRoutingMode;
 import org.apache.pulsar.client.api.ProducerAccessMode;
 import org.apache.pulsar.client.api.ProducerCryptoFailureAction;
 import org.apache.pulsar.client.api.PulsarClient;
+import org.apache.pulsar.client.api.PulsarClientException.ProducerQueueIsFullError;
 import org.apache.pulsar.client.api.Schema;
 import org.apache.pulsar.client.impl.MessageIdImpl;
 import org.apache.pulsar.client.impl.ProducerBase;
@@ -53,18 +56,19 @@ import org.apache.pulsar.client.impl.TypedMessageBuilderImpl;
 import org.apache.pulsar.client.impl.conf.ProducerConfigurationData;
 import org.apache.pulsar.client.internal.DefaultImplementation;
 import org.apache.pulsar.reactive.client.adapter.AdaptedReactivePulsarClientFactory;
+import org.apache.pulsar.reactive.client.api.MessageSendResult;
 import org.apache.pulsar.reactive.client.api.MessageSpec;
 import org.apache.pulsar.reactive.client.api.ReactiveMessageSender;
 import org.apache.pulsar.reactive.client.api.ReactiveMessageSenderBuilder;
 import org.apache.pulsar.reactive.client.api.ReactiveMessageSenderCache;
+import org.apache.pulsar.reactive.client.api.ReactiveMessageSendingException;
 import org.apache.pulsar.reactive.client.internal.api.InternalMessageSpec;
+import org.assertj.core.api.InstanceOfAssertFactories;
 import org.junit.jupiter.api.Test;
 import org.junit.jupiter.params.ParameterizedTest;
 import org.junit.jupiter.params.provider.Arguments;
 import org.junit.jupiter.params.provider.CsvSource;
 import org.junit.jupiter.params.provider.MethodSource;
-import org.mockito.InOrder;
-import org.mockito.Mockito;
 import reactor.core.publisher.Flux;
 import reactor.test.StepVerifier;
 
@@ -163,36 +167,132 @@ class AdaptedReactiveMessageSenderTest {
 	}
 
 	@Test
-	void sendMany() throws Exception {
+	void sendOnePulsarException() throws Exception {
 		PulsarClientImpl pulsarClient = spy(
 				(PulsarClientImpl) PulsarClient.builder().serviceUrl("http://dummy").build());
 
 		ProducerBase<String> producer = mock(ProducerBase.class);
 		doReturn(CompletableFuture.completedFuture(null)).when(producer).closeAsync();
-		TypedMessageBuilderImpl<String> typedMessageBuilder1 = spy(
-				new TypedMessageBuilderImpl<>(producer, Schema.STRING));
-		doReturn(CompletableFuture.completedFuture(MessageId.earliest)).when(typedMessageBuilder1).sendAsync();
-		TypedMessageBuilderImpl<String> typedMessageBuilder2 = spy(
-				new TypedMessageBuilderImpl<>(producer, Schema.STRING));
-		doReturn(CompletableFuture.completedFuture(MessageId.latest)).when(typedMessageBuilder2).sendAsync();
 
-		doReturn(typedMessageBuilder1, typedMessageBuilder2).when(producer).newMessage();
+		given(producer.newMessage()).willAnswer((__) -> {
+			TypedMessageBuilderImpl<String> typedMessageBuilder = spy(
+					new TypedMessageBuilderImpl<>(producer, Schema.STRING));
+			given(typedMessageBuilder.sendAsync()).willAnswer((___) -> {
+				CompletableFuture<MessageId> failed = new CompletableFuture<>();
+				failed.completeExceptionally(new ProducerQueueIsFullError("Queue is full"));
+				return failed;
+			});
+			return typedMessageBuilder;
+		});
+
 		doReturn(CompletableFuture.completedFuture(producer)).when(pulsarClient).createProducerAsync(any(),
 				eq(Schema.STRING), isNull());
 
 		ReactiveMessageSender<String> reactiveSender = AdaptedReactivePulsarClientFactory.create(pulsarClient)
 				.messageSender(Schema.STRING).topic("my-topic").build();
 
-		Flux<MessageSpec<String>> messageSpecs = Flux.just(MessageSpec.of("test1"), MessageSpec.of("test2"));
-		StepVerifier.create(reactiveSender.sendMany(messageSpecs)).expectNext(MessageId.earliest)
-				.expectNext(MessageId.latest).verifyComplete();
+		StepVerifier.create(reactiveSender.sendOne(MessageSpec.of("test1")))
+				// the original exception should be returned without wrapping it in
+				// ReactiveMessageSendingException
+				.expectError(ProducerQueueIsFullError.class).verify();
+	}
 
-		verify(pulsarClient).createProducerAsync(any(), any(), isNull());
-		InOrder inOrder = Mockito.inOrder(typedMessageBuilder1, typedMessageBuilder2);
-		inOrder.verify(typedMessageBuilder1).value("test1");
-		inOrder.verify(typedMessageBuilder1).sendAsync();
-		inOrder.verify(typedMessageBuilder2).value("test2");
-		inOrder.verify(typedMessageBuilder2).sendAsync();
+	@Test
+	void sendManyStopOnError() throws Exception {
+		PulsarClientImpl pulsarClient = spy(
+				(PulsarClientImpl) PulsarClient.builder().serviceUrl("http://dummy").build());
+
+		ProducerBase<String> producer = mock(ProducerBase.class);
+		doReturn(CompletableFuture.completedFuture(null)).when(producer).closeAsync();
+
+		AtomicInteger entryId = new AtomicInteger();
+		List<MessageId> messageIds = new CopyOnWriteArrayList<>();
+		given(producer.newMessage()).willAnswer((__) -> {
+			TypedMessageBuilderImpl<String> typedMessageBuilder = spy(
+					new TypedMessageBuilderImpl<>(producer, Schema.STRING));
+			given(typedMessageBuilder.sendAsync()).willAnswer((___) -> {
+				if (entryId.get() == 1) {
+					CompletableFuture<MessageId> failed = new CompletableFuture<>();
+					failed.completeExceptionally(new ProducerQueueIsFullError("Queue is full"));
+					return failed;
+				}
+				MessageId messageId = DefaultImplementation.getDefaultImplementation().newMessageId(1,
+						entryId.incrementAndGet(), 1);
+				messageIds.add(messageId);
+				return CompletableFuture.completedFuture(messageId);
+			});
+			return typedMessageBuilder;
+		});
+
+		doReturn(CompletableFuture.completedFuture(producer)).when(pulsarClient).createProducerAsync(any(),
+				eq(Schema.STRING), isNull());
+
+		ReactiveMessageSender<String> reactiveSender = AdaptedReactivePulsarClientFactory.create(pulsarClient)
+				.messageSender(Schema.STRING).topic("my-topic").stopOnError(true).build();
+
+		MessageSpec<String> failingMessage = MessageSpec.builder("test2").correlationMetadata("my-context").build();
+		Flux<MessageSpec<String>> messageSpecs = Flux.just(MessageSpec.of("test1"), failingMessage);
+		StepVerifier.create(reactiveSender.sendMany(messageSpecs))
+				.assertNext((next) -> assertThat(next.getMessageId()).isEqualTo(messageIds.get(0)))
+				.verifyErrorSatisfies((throwable) -> assertThat(throwable)
+						.asInstanceOf(InstanceOfAssertFactories.type(ReactiveMessageSendingException.class))
+						.satisfies((rmse) -> assertThat(rmse.getCause()).isInstanceOf(ProducerQueueIsFullError.class))
+						.satisfies((rmse) -> assertThat(rmse.getMessageSpec()).isSameAs(failingMessage))
+						.satisfies((rmse) -> assertThat((String) rmse.getCorrelationMetadata()).isEqualTo("my-context"))
+						.satisfies(
+								(rmse) -> assertThat(rmse.toString()).contains("correlation metadata={my-context}")));
+	}
+
+	@Test
+	void sendMany() throws Exception {
+		PulsarClientImpl pulsarClient = spy(
+				(PulsarClientImpl) PulsarClient.builder().serviceUrl("http://dummy").build());
+
+		ProducerBase<String> producer = mock(ProducerBase.class);
+		doReturn(CompletableFuture.completedFuture(null)).when(producer).closeAsync();
+
+		AtomicInteger entryId = new AtomicInteger();
+		List<MessageId> messageIds = new CopyOnWriteArrayList<>();
+		given(producer.newMessage()).willAnswer((__) -> {
+			TypedMessageBuilderImpl<String> typedMessageBuilder = spy(
+					new TypedMessageBuilderImpl<>(producer, Schema.STRING));
+			given(typedMessageBuilder.sendAsync()).willAnswer((___) -> {
+				if (entryId.get() == 2) {
+					CompletableFuture<MessageId> failed = new CompletableFuture<>();
+					failed.completeExceptionally(new ProducerQueueIsFullError("Queue is full"));
+					return failed;
+				}
+				MessageId messageId = DefaultImplementation.getDefaultImplementation().newMessageId(1,
+						entryId.incrementAndGet(), 1);
+				messageIds.add(messageId);
+				return CompletableFuture.completedFuture(messageId);
+			});
+			return typedMessageBuilder;
+		});
+
+		doReturn(CompletableFuture.completedFuture(producer)).when(pulsarClient).createProducerAsync(any(),
+				eq(Schema.STRING), isNull());
+
+		ReactiveMessageSender<String> reactiveSender = AdaptedReactivePulsarClientFactory.create(pulsarClient)
+				.messageSender(Schema.STRING).topic("my-topic").build();
+
+		MessageSpec<String> messageSpec1 = MessageSpec.of("test1");
+		Flux<MessageSpec<String>> keysAndMessageSpecs = Flux.just(messageSpec1,
+				MessageSpec.builder("test2").correlationMetadata(456).build(),
+				MessageSpec.builder("test3").correlationMetadata(789).build());
+		StepVerifier.create(reactiveSender.sendMany(keysAndMessageSpecs)).assertNext((next) -> {
+			assertThat(next.getMessageId()).isEqualTo(messageIds.get(0));
+			assertThat(next.getMessageSpec()).isEqualTo(messageSpec1);
+			assertThat(next.getException()).isNull();
+		}).assertNext((next) -> {
+			assertThat(next.getMessageId()).isEqualTo(messageIds.get(1));
+			assertThat((int) next.getCorrelationMetadata()).isEqualTo(456);
+			assertThat(next.getException()).isNull();
+		}).assertNext((next) -> {
+			assertThat(next.getMessageId()).isNull();
+			assertThat((int) next.getCorrelationMetadata()).isEqualTo(789);
+			assertThat(next.getException()).isInstanceOf(ProducerQueueIsFullError.class);
+		}).verifyComplete();
 	}
 
 	@ParameterizedTest
@@ -344,8 +444,10 @@ class AdaptedReactiveMessageSenderTest {
 	@ParameterizedTest
 	@CsvSource({ "7,100", "13,100", "37,500", "51,1000" })
 	void maxInFlightUsingSendMany(int maxInflight, int maxElements) throws Exception {
-		doTestMaxInFlight((reactiveSender, inputFlux) -> inputFlux.window(3).flatMap(
-				(subFlux) -> subFlux.map((i) -> MessageSpec.of(String.valueOf(i))).as(reactiveSender::sendMany), 100),
+		doTestMaxInFlight(
+				(reactiveSender, inputFlux) -> inputFlux.window(3)
+						.flatMap((subFlux) -> subFlux.map((i) -> MessageSpec.of(String.valueOf(i)))
+								.as(reactiveSender::sendMany).map(MessageSendResult::getMessageId), 100),
 				maxInflight, maxElements);
 	}
 
